
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.solver.Solution;
import org.chocosolver.solver.search.strategy.selectors.variables.ImpactBased;
import org.chocosolver.solver.search.strategy.strategy.StrategiesSequencer;

public class Bombe {
		
	
		static public void BombeEnigmaBreak(EnigmaEmulator machine, String code, String crib, int[] key) {
			Alphabet alphabet = machine.get_alphabet();
			char[] code_char = code.toCharArray();
			char[] crib_char = crib.toCharArray();
			Model model = new Model();

			int[][] frotors = machine.get_forward_rotors();
			int[][] brotors = machine.get_backward_rotors();
			int N = alphabet.size();
			int M = frotors.length;
			int K = code_char.length;
			int[] reflector = machine.get_reflector();


			// IntVar[] key = model.intVarArray("key", M, 0, N-1);
			IntVar[] plug_board = model.intVarArray("plug", N, 0, N-1);
			BoolVar[] loop = model.boolVarArray("loop", N);

			// IntVar[] plaintext = model.intVarArray("plain", code_char.length, 0, N-1);
			IntVar[][] text = new IntVar[2*M+4][];
			text[0] = model.intVarArray("plain", K, 0, N-1);
			text[1] = model.intVarArray("switched", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[j+1] = model.intVarArray(String.format("forward[%d]",j), K, 0, N-1);
			}
			text[M+2] = model.intVarArray("reflected", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[M+2+j] = model.intVarArray(String.format("backward[%d]",j), K, 0, N-1);
			}
			text[2*M+3] = model.intVarArray("cipher", K, 0, N-1);
			
			model.allDifferent(plug_board).post();
			for(int i=0; i<N; ++i) {
					model.reifyXeqC(plug_board[i], i, loop[i]);
			}
			model.sum(loop, "=", 6);
			


			for(int i=0; i<K; ++i) {
					model.element( text[1][i], plug_board, text[0][i], 0 ).post();
					model.element( text[0][i], plug_board, text[1][i], 0 ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[j+1][i].add(key[j]).add((i/((int)(Math.pow(N,j))))).mod(N).intVar();
							model.element( text[j+2][i], frotors[j], index ).post();
					}
					model.element( text[M+2][i], reflector, text[M+1][i] ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[M+3+j][i].add(key[(2-j)]).add((i/((int)(Math.pow(N,(2-j)))))).mod(N).intVar();
							model.element( text[M+2+j][i], frotors[2-j], index ).post();
					}
					model.element( text[2*M+3][i], plug_board, text[2*M+2][i], 0 ).post();
					model.element( text[2*M+2][i], plug_board, text[2*M+3][i], 0 ).post();
					
					model.arithm( text[2*M+3][i], "=", alphabet.get_digit(code_char[i]) ).post();
			}

			for(int i=0; i<crib_char.length; ++i) {
					model.arithm( text[0][i], "=", alphabet.get_digit(crib_char[i]) ).post();
			}
			
      for(int i=0; i<crib_char.length; ++i) {
          for(int j=i+1; j<crib_char.length; ++j) {
              if(crib_char[i] == crib_char[j])
                  model.arithm( text[1][i], "=", text[1][j] ).post();
              else
                  model.arithm( text[1][i], "!=", text[1][j] ).post();
              
              if(code_char[i] == code_char[j])
                  model.arithm( text[2*M+2][j], "=", text[2*M+2][i] ).post();
              else
                  model.arithm( text[2*M+2][j], "!=", text[2*M+2][i] ).post();
              
              if(crib_char[i] == code_char[j])
                  model.arithm( text[1][i], "=", text[2*M+2][j] ).post();
              else
                  model.arithm( text[1][i], "!=", text[2*M+2][j] ).post();
              
              if(code_char[i] == crib_char[j])
                  model.arithm( text[2*M+2][i], "=", text[1][j] ).post();
              else
                  model.arithm( text[2*M+2][i], "!=", text[1][j] ).post();
					}
			}

			Solver solver = model.getSolver();
			
			// solver.setSearch(new ImpactBased(plug_board, false), new ImpactBased(text[0], false));
			// solver.showStatisticsDuringResolution(1000);
			// solver.showShortStatistics();
			
			Solution solution = solver.findSolution();
			while(solution != null) {
					alphabet.print(text[0]);
					System.out.println();
					
					solution = solver.findSolution();
			}

		}
		
		
		public static void tryKeys(EnigmaEmulator machine, String ciphertext, String crib) {
			
			int ITER = -1;
			
			
		    Alphabet alphabet = machine.get_alphabet();
				int M = machine.num_rotors();
				int N = alphabet.size();
				int K = crib.length();
				
				int[] id = new int[N];
				for(int i=0; i<N; ++i) id[i] = i;

				char[] crib_char = crib.toCharArray();
				char[] code_char = ciphertext.substring(0,K).toCharArray();
			
				assert code_char.length == K;

				boolean[] in_crib = new boolean[N];
				Arrays.fill(in_crib, false);
				for(char c : crib_char) in_crib[alphabet.get_digit(c)] = true;
				for(char c : code_char) in_crib[alphabet.get_digit(c)] = true;

				int[][] neighbor_list = new int[N][];
				int[][] menu = new int[N][N];
				for(int i=0; i<N; ++i) {
						Arrays.fill(menu[i], -1);
				}
			
				for(int i=0; i<K; ++i) {
						int x = alphabet.get_digit(crib_char[i]);
						int y = alphabet.get_digit(code_char[i]);
						menu[x][y] = i;
						menu[y][x] = i;
				}
			
				for(int x=0; x<N; ++x) {	
						List<Integer> neighbors = new ArrayList<>();
						for(int y=0; y<N; ++y) if(menu[x][y] >= 0) neighbors.add(y);
					
						neighbor_list[x] = new int[neighbors.size()];
						for(int i=0; i<neighbors.size(); ++i) neighbor_list[x][i] = neighbors.get(i);
				}		
				
        boolean[][] forbidden = new boolean[N][N];
        int[] possible = new int[N];
				int[] stecker = new int[N];

				int[] stack = new int[2*N*N];
				
				int max_iterations = (int)Math.pow(N,M);
				int[] key = new int[M];
				int[] ikey = new int[M];
				
				
				for(int i=0; i<N; ++i) {
					System.out.print( alphabet.get_char(i) + " ->" );
					for(int j=0; j<neighbor_list[i].length; ++j) {
						int n = neighbor_list[i][j];
						System.out.print( " " + alphabet.get_char(n) + " (" + menu[i][n] + ")" );
					}
					System.out.println( );
				}
				
				// if(true)
				// 	return;
				
				int nkeys = 0;
				for(int iteration = 0; iteration<max_iterations; ++iteration) {
								        if(iteration % 100 == 0) {
								            System.out.printf("\b\b\b\b\b\b\b\b\b%-5d %2d%%", nkeys, (iteration*100)/max_iterations);
								System.out.flush();
						}

						// System.out.print( nkeys + " " + iteration );
						for(int i=0; i<M; ++i) {
								key[i] = (iteration/(int)Math.pow(N,i)) % N;
								// System.out.print( " " + key[i] );
						}
						// System.out.println();
						
						
											
						boolean support = true;
						int diagonal = N;
						for(int i=0; i<N; ++i) {
								Arrays.fill(forbidden[i], false);
								possible[i] = N;
						}
						
						for(int s=0; s<N; ++s) {
								if(in_crib[s]) {
										for(int t=0; t<N; ++t) {
												int top = 0;
												stack[top++] = s;
												stack[top++] = t;
												stack[top++] = t;
												stack[top++] = s;
												
												
												
												int pos = 0;
												Arrays.fill(stecker, -1);
												
												while( pos < top && !forbidden[t][s] ) {
														int x = stack[pos++];
														int y = stack[pos++];
																												
														if( !forbidden[x][y] ) {
				                        stecker[x] = y;
				                        stecker[y] = x;
																for(int j=0; j<neighbor_list[x].length && support; ++j) {
																		int cx = neighbor_list[x][j];
																		
																		
																		for(int k=0; k<M; ++k) {
																				ikey[k] = (key[k] + menu[x][cx]/(int)Math.pow(N,k)) % N;
																		}
																		machine.set_key(ikey);																		
																		
																		int ey = machine.encryption(y);
																		
				                            if(stecker[ey]<0 && stecker[cx]<0) {
				                                stecker[ey] = cx;
				                                stecker[cx] = ey;
																				stack[top++] = cx;
																				stack[top++] = ey;
				                                if(ey != cx) {
																						stack[top++] = ey;
																						stack[top++] = cx;
																				}
				                            } else if(stecker[cx] != ey) {

																				stack[top++] = cx;
																				stack[top++] = ey;
				                                if(ey != cx) {
																						stack[top++] = ey;
																						stack[top++] = cx;
																				}
																				while(top>0 && support) {
																						int c1 = stack[--top];
																						int c2 = stack[--top];
																						if(!forbidden[c1][c2]) {
																								forbidden[c1][c2] = true;
																								if(--possible[c1] == 0) support = false;	
																						}
																						if(c1 == c2) 
																								if(--diagonal < 6) support = false;
																						
																				}
																				break;
																		}
																}
														}
												}
										}
								}
						}
		        if(support) {
		            ++nkeys;            
		            System.out.print("\b\b\b\b\b\b\b\b\bSTOP");
								for( int k : key ) {
										System.out.print( " " + k );
								}
								System.out.printf(" @%d\n", iteration);
		            BombeEnigmaBreak(machine, ciphertext, crib, key);
						} 
				}
				System.out.println();
		}
		
		

    public static void main(String[] args) {
		    Alphabet alphabet = new Alphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    
  	
		    int[][] perm = {{0, 9, 20, 19, 7, 6, 4, 11, 3, 14, 22, 8, 17, 15, 2, 16, 23, 18, 10, 1, 12, 21, 24, 13, 5, 25}, 
		            				{20, 15, 5, 12, 25, 14, 16, 2, 22, 21, 7, 9, 1, 24, 4, 19, 3, 11, 0, 18, 13, 10, 8, 17, 23, 6}, 
		            				{13, 0, 2, 4, 14, 19, 16, 22, 25, 1, 7, 9, 17, 10, 3, 21, 15, 5, 24, 11, 12, 8, 20, 23, 18, 6}};
		    int[] reflector = {21, 13, 12, 14, 5, 4, 18, 25, 22, 23, 17, 24, 2, 1, 3, 16, 15, 10, 6, 20, 19, 0, 8, 9, 11, 7};
				int[] plug_board = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25};


				String cipher = "CAUBOJJOUTCTVHXQFJZUBZQCNVIGFBUTJXSSZJRIWTPEKDJPZCVNEPQMCABQVZOYJRJJACYVJJGWFZNZTFLCEIGGHCNBXYGMUQHUYXVHTDDQLVCOKTHXCZSVSPGWTGAAGQVWKWOD";
    
				EnigmaEmulator machine = new EnigmaEmulator(alphabet, perm, plug_board, reflector);
		
				Bombe.tryKeys(machine, cipher, "THEFIRSTONEWHOCAN");
		
    }
}

// CAUBOJJOUTCTVHXQFJZUBZQCNVIGFBUTJXSSZJRIWTPEKDJPZCVNEPQMCABQVZOYJRJJACYVJJGWFZNZTFLCEIGGHCNBXYGMUQHUYXVHTDDQLVCOKTHXCZSVSPGWTGAAGQVWKWOD
   // THEFIRSTONEWHOCAN

