import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.Solution;


public class CommercialEnigmaBreaker {
	
	
		static public void singleRotorEmulate(Alphabet alphabet, int[] permutation, int length) {
						
				Model model = new Model();	
			
				int N = alphabet.size();
			
				IntVar[] plaintext = model.intVarArray("plain", length, 0, N-1);
				IntVar[] ciphertext = model.intVarArray("cipher", length, 0, N-1);
				IntVar key = model.intVar("key", 0, N-1);
			
				for(int i=0; i<length; ++i) {
						IntVar index = plaintext[i].add(key).add(i).mod(N).intVar();
						model.element( ciphertext[i], permutation, index ).post();
				}
			
			
				Solver solver = model.getSolver();
				solver.findSolution();
				for(int k=0; k<1000; ++k) {
						alphabet.print(plaintext);
						System.out.print(" -[");
						System.out.print(key.getValue());
						System.out.print("]-> ");
						alphabet.print(ciphertext);
						System.out.println();
								
						solver.findSolution();
				}
		
		}
		
		static public void singleRotorBreak(Alphabet alphabet, int[] rotor, String code, String crib) {
				
				char[] code_char = code.toCharArray();
				char[] crib_char = crib.toCharArray();		
				Model model = new Model();
			
				int N = alphabet.size();
				int K = code_char.length;
			
				IntVar[] plaintext = model.intVarArray("plain", K, 0, N-1);
				IntVar[] ciphertext = model.intVarArray("cipher", K, 0, N-1);
				IntVar key = model.intVar("key", 0, N-1);
			
				for(int i=0; i<K; ++i) {
						IntVar index = plaintext[i].add(key).add(i).mod(N).intVar();
						model.element( ciphertext[i], rotor, index ).post();
						model.arithm( ciphertext[i], "=", alphabet.get_digit(code_char[i]) ).post();
				}
				
				
				
				for(int i=0; i<crib_char.length; ++i) {
						model.arithm( plaintext[i], "=", alphabet.get_digit(crib_char[i]) ).post();
				}
				
			
				Solver solver = model.getSolver();
				Solution solution = solver.findSolution();
				while(solution != null) {
						alphabet.print(plaintext);
						System.out.print(" -[");
						System.out.print(key.getValue());
						System.out.print("]-> ");
						alphabet.print(ciphertext);
						System.out.println();
									
						solution = solver.findSolution();
				}
		
		}
		
		
		static public void multipleRotorsBreak(Alphabet alphabet, int[][] rotors, String code, String crib) {
			
			char[] code_char = code.toCharArray();
			char[] crib_char = crib.toCharArray();		
			Model model = new Model();
		
			int N = alphabet.size();
			int K = code_char.length;
			int M = rotors.length;
		
			// IntVar[] plaintext = model.intVarArray("plain", K, 0, N-1);
			IntVar[][] text = new IntVar[M+1][];
			text[0] = model.intVarArray("plain", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[j] = model.intVarArray(String.format("cipher[%d]",j), K, 0, N-1);
			}
			IntVar[] key = model.intVarArray("key", M, 0, N-1);
		
			for(int i=0; i<K; ++i) {
					for(int j=0; j<M; ++j) {
							IntVar index = text[j][i].add(key[j]).add((i/((int)(Math.pow(N,j))))).mod(N).intVar();
							model.element( text[j+1][i], rotors[j], index ).post();
					}
					model.arithm( text[M][i], "=", alphabet.get_digit(code_char[i]) ).post();
			}
			
			for(int i=0; i<crib_char.length; ++i) {
					model.arithm( text[0][i], "=", alphabet.get_digit(crib_char[i]) ).post();
			}
			
		
			Solver solver = model.getSolver();
			Solution solution = solver.findSolution();
			while(solution != null) {
					alphabet.print(text[0]);
					System.out.print(" -[");
					for(int i=0; i<M; ++i) {
							System.out.print(" " + key[i].getValue());
					}
					System.out.print(" ]-> ");
					alphabet.print(text[M]);
					System.out.println();
			
					solution = solver.findSolution();
			}

		}
		
		
		
		static public void commercialEnigmaEncode(EnigmaEmulator machine, String code, int[] key) {
			Alphabet alphabet = machine.get_alphabet();
			char[] code_char = code.toCharArray();
			Model model = new Model();
		
			int[][] frotors = machine.get_forward_rotors();
			int[][] brotors = machine.get_backward_rotors();
			int N = alphabet.size();
			int M = frotors.length;
			int K = code_char.length;
			int[] reflector = machine.get_reflector();
		
			// IntVar[] plaintext = model.intVarArray("plain", K, 0, N-1);
			IntVar[][] text = new IntVar[2*M+2][];
			text[0] = model.intVarArray("plain", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[j] = model.intVarArray(String.format("forward[%d]",j), K, 0, N-1);
			}
			text[M+1] = model.intVarArray("reflected", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[M+1+j] = model.intVarArray(String.format("backward[%d]",j), K, 0, N-1);
			}
			
			for(int i=0; i<K; ++i) {
					for(int j=0; j<M; ++j) {
							IntVar index = text[j][i].add(key[j]+(i/((int)(Math.pow(N,j))))).mod(N).intVar();
							model.element( text[j+1][i], frotors[j], index ).post();
					}
					model.element( text[M+1][i], reflector, text[M][i] ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[M+2+j][i].add(key[(2-j)]+(i/((int)(Math.pow(N,(2-j)))))).mod(N).intVar();
							model.element( text[M+1+j][i], frotors[2-j], index ).post();
					}
					model.arithm( text[0][i], "=", alphabet.get_digit(code_char[i]) ).post();
			}
			
		
			Solver solver = model.getSolver();
			Solution solution = solver.findSolution();
			while(solution != null) {
					alphabet.print(text[2*M+1]);
					System.out.println();
					solution = solver.findSolution();
			}

		}
		
		

		static public void commercialEnigmaBreak(EnigmaEmulator machine, String code, String crib) {
			Alphabet alphabet = machine.get_alphabet();
			char[] code_char = code.toCharArray();
			char[] crib_char = crib.toCharArray();
			Model model = new Model();
			
			int[][] frotors = machine.get_forward_rotors();
			int[][] brotors = machine.get_backward_rotors();
			int N = alphabet.size();
			int M = frotors.length;
			int K = code_char.length;
			int[] reflector = machine.get_reflector();
			
			
			IntVar[] key = model.intVarArray("key", M, 0, N-1);
		
			// IntVar[] plaintext = model.intVarArray("plain", K, 0, N-1);
			IntVar[][] text = new IntVar[2*M+2][];
			text[0] = model.intVarArray("plain", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[j] = model.intVarArray(String.format("forward[%d]",j), K, 0, N-1);
			}
			text[M+1] = model.intVarArray("reflected", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[M+1+j] = model.intVarArray(String.format("backward[%d]",j), K, 0, N-1);
			}
			
			for(int i=0; i<K; ++i) {
					for(int j=0; j<M; ++j) {
							IntVar index = text[j][i].add(key[j]).add((i/((int)(Math.pow(N,j))))).mod(N).intVar();
							model.element( text[j+1][i], frotors[j], index ).post();
					}
					model.element( text[M+1][i], reflector, text[M][i] ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[M+2+j][i].add(key[(2-j)]).add((i/((int)(Math.pow(N,(2-j)))))).mod(N).intVar();
							model.element( text[M+1+j][i], frotors[2-j], index ).post();
					}
					model.arithm( text[2*M+1][i], "=", alphabet.get_digit(code_char[i]) ).post();
			}
			
			for(int i=0; i<crib_char.length; ++i) {
					model.arithm( text[0][i], "=", alphabet.get_digit(crib_char[i]) ).post();
			}
			
		
			Solver solver = model.getSolver();
			Solution solution = solver.findSolution();
			while(solution != null) {
					alphabet.print(text[0]);
					System.out.println();
					solution = solver.findSolution();
			}

		}
		
		
		

    public static void main(String[] args) {
				Alphabet alphabet = new Alphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
			
				int[][] perm = new int[][]{
					{15, 14, 5, 22, 24, 13, 25, 12, 18, 9, 1, 23, 20, 16, 17, 21, 7, 2, 3, 11, 4, 8, 6, 19, 0, 10},
					{21, 17, 13, 6, 0, 16, 10, 20, 3, 9, 14, 19, 4, 7, 2, 8, 12, 5, 22, 23, 24, 18, 25, 11, 15, 1},
					{8, 2, 16, 25, 15, 7, 1, 4, 22, 9, 17, 11, 24, 12, 5, 18, 20, 13, 6, 19, 0, 3, 10, 23, 21, 14}
				};
			
				int[] reflector = {10, 18, 16, 6, 15, 19, 3, 12, 20, 23, 0, 14, 7, 22, 11, 4, 2, 24, 1, 5, 8, 25, 13, 9, 17, 21};
				int[] plug_board = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25};
				
				System.out.println("EXERCICE 1\n");
				
				singleRotorEmulate(alphabet, perm[0], 3);
			

		    System.out.println("\n\nEXERCICE 2\n");
		    String cipher0 = "LCKPYBKOUMPVHKZNBITLIKUXOVN";

				/*
		    The ciphertext above was gereated as follows:

		    k = ?
		    text = 'H??????????????????????????'
		    print ''.join([alphabet[perm1[(digit[t]+i+k) % N]] for i,t in enumerate(text)])
		    */

		    singleRotorBreak(alphabet, perm[0], cipher0, "H");
				
				
		    System.out.println("\n\nEXERCICE 3\n");
		    String cipher1 = "YTHXGIPIBUZYCCQSLPELWDRAOZXOSGOHCZNOLRGYXFYZTWFMZIBDBUVOU";

		    /*
		    The ciphertext above was gereated as follows:

		    text = 'TH???????????????????????????????????????????????????????'
		    rotors[0].set_key(12)
		    rotors[1].set_key(7)

		    cipher = ''
		    for i,c in enumerate(text):
		        c = digit[c]
		        for rotor in rotors:
		            c = rotor.encrypt_forward(c)
		        j = 0
		        while j<len(rotors) and rotors[j].step():
		            j += 1
		        cipher += alphabet[c]

		    print cipher
		    */

		    multipleRotorsBreak(alphabet, new int[][]{perm[0]}, cipher0, "H");
		    System.out.println();
		    multipleRotorsBreak(alphabet, new int[][]{perm[0], perm[1]}, cipher1, "TH");
				
				
		    System.out.println("\n\nEXERCICE 4");
		    EnigmaEmulator machine = new EnigmaEmulator(alphabet, perm, plug_board, reflector);

		    String T = "TESTFORMYENIGMACONSTRAINTPROGRAMMINGMODEL";

		    for(int i=0; i<1; ++i) {
		        int[] key = new int[]{3,4,5};

		        System.out.println();

		        machine.set_key(key);
						String ciphertext = machine.encrypt(T);
		        System.out.println("Enigma  cipher: "+ciphertext);
						System.out.print("CPmodel cipher: ");
		        commercialEnigmaEncode(machine, T, key);

		        System.out.println();
						
		        machine.set_key(key);
						String plaintext = machine.encrypt(ciphertext);
		        System.out.println("Enigma  plaintext: "+plaintext);
						System.out.print("CPmodel plaintext: ");
		        commercialEnigmaEncode(machine, ciphertext, key);
				}
				
		    System.out.println("\n\nEXERCICE 5\n");
		    String cipher2 = "SLIAMQPKAYMDNZPPJUWWWTNFCJZEHRNQOEANJGVBMHHDZBXMACJRZYPDXJUSWTGOPCJQUJCZVAMEDUPAIMPCXP";
 
		    /*
		    The code above was gereated as follows:

		    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=range(N), reflector=reflector)
        
		    enigma.set_key([?,?,?])
		    cipher2 = enigma.encrypt('WELL??????????????????????????????????????????????????????????????????????????????????')

		    print cipher2
		    */
    
    
		    commercialEnigmaBreak(machine, cipher2, "WE");

				
    }
}
