import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.solver.Solution;
import org.chocosolver.solver.search.strategy.selectors.variables.ImpactBased;
import org.chocosolver.solver.search.strategy.strategy.StrategiesSequencer;

public class MilitaryEnigmaBreaker {
		
	
		static public void militaryEnigmaBreak(EnigmaEmulator machine, int nplugs, String code, String crib) {
			Alphabet alphabet = machine.get_alphabet();
			char[] code_char = code.toCharArray();
			char[] crib_char = crib.toCharArray();
			Model model = new Model();

			int[][] frotors = machine.get_forward_rotors();
			int[][] brotors = machine.get_backward_rotors();
			int N = alphabet.size();
			int M = frotors.length;
			int K = code_char.length;
			int[] reflector = machine.get_reflector();


			IntVar[] key = model.intVarArray("key", M, 0, N-1);
			IntVar[] plug_board = model.intVarArray("plug", N, 0, N-1);
			BoolVar[] loop = model.boolVarArray("loop", N);

			// IntVar[] plaintext = model.intVarArray("plain", code_char.length, 0, N-1);
			IntVar[][] text = new IntVar[2*M+4][];
			text[0] = model.intVarArray("plain", K, 0, N-1);
			text[1] = model.intVarArray("switched", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[j+1] = model.intVarArray(String.format("forward[%d]",j), K, 0, N-1);
			}
			text[M+2] = model.intVarArray("reflected", K, 0, N-1);
			for(int j=1; j<=M; ++j) {
					text[M+2+j] = model.intVarArray(String.format("backward[%d]",j), K, 0, N-1);
			}
			text[2*M+3] = model.intVarArray("cipher", K, 0, N-1);
			
			model.allDifferent(plug_board).post();
			for(int i=0; i<N; ++i) {
					model.reifyXeqC(plug_board[i], i, loop[i]);
			}
			model.sum(loop, "=", N-2*nplugs);
			


			for(int i=0; i<K; ++i) {
					model.element( text[1][i], plug_board, text[0][i], 0 ).post();
					// model.element( text[0][i], plug_board, text[1][i], 0 ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[j+1][i].add(key[j]).add((i/((int)(Math.pow(N,j))))).mod(N).intVar();
							model.element( text[j+2][i], frotors[j], index ).post();
					}
					model.element( text[M+2][i], reflector, text[M+1][i] ).post();
					for(int j=0; j<M; ++j) {
							IntVar index = text[M+3+j][i].add(key[(2-j)]).add((i/((int)(Math.pow(N,(2-j)))))).mod(N).intVar();
							model.element( text[M+2+j][i], frotors[2-j], index ).post();
					}
					model.element( text[2*M+3][i], plug_board, text[2*M+2][i], 0 ).post();
					// model.element( text[2*M+2][i], plug_board, text[2*M+3][i], 0 ).post();
					model.arithm( text[2*M+3][i], "=", alphabet.get_digit(code_char[i]) ).post();
			}

			for(int i=0; i<crib_char.length; ++i) {
					model.arithm( text[0][i], "=", alphabet.get_digit(crib_char[i]) ).post();
			}
			
      for(int i=0; i<crib_char.length; ++i) {
          for(int j=i+1; j<crib_char.length; ++j) {
              if(crib_char[i] == crib_char[j])
                  model.arithm( text[1][i], "=", text[1][j] ).post();
              else
                  model.arithm( text[1][i], "!=", text[1][j] ).post();
              
              if(code_char[i] == code_char[j])
                  model.arithm( text[2*M+2][j], "=", text[2*M+2][i] ).post();
              else
                  model.arithm( text[2*M+2][j], "!=", text[2*M+2][i] ).post();
              
              if(crib_char[i] == code_char[j])
                  model.arithm( text[1][i], "=", text[2*M+2][j] ).post();
              else
                  model.arithm( text[1][i], "!=", text[2*M+2][j] ).post();
              
              if(code_char[i] == crib_char[j])
                  model.arithm( text[2*M+2][i], "=", text[1][j] ).post();
              else
                  model.arithm( text[2*M+2][i], "!=", text[1][j] ).post();
					}
			}

			Solver solver = model.getSolver();
			
			solver.setSearch(new ImpactBased(key, false), new ImpactBased(plug_board, false), new ImpactBased(text[0], false));
			solver.showStatisticsDuringResolution(1000);
			solver.showShortStatistics();
			
			Solution solution = solver.findSolution();
			while(solution != null) {
					alphabet.print(text[0]);
					System.out.println();
					
					solution = solver.findSolution();
			}

		}
		
		

    public static void main(String[] args) {
	    Alphabet alphabet = new Alphabet("ACDEGHILMNORSTWY");
    
    
	    int[][] perm = new int[][]{{13, 9, 8, 10, 7, 15, 12, 14, 1, 5, 2, 4, 3, 11, 0, 6}, 
	            {1, 5, 2, 3, 7, 11, 8, 13, 10, 4, 14, 6, 12, 15, 0, 9}, 
	            {9, 13, 1, 10, 14, 0, 3, 12, 8, 11, 4, 7, 5, 2, 6, 15}};
	    int[] reflector = new int[]{4, 13, 5, 7, 0, 2, 15, 3, 9, 8, 12, 14, 10, 1, 11, 6};
			int[] plug_board = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}; // this is not the real plug board!
    

	    System.out.println("\n\nEXERCICE 1\n");
	    String cipher2 = "RCSDRLWISTHIYMRLTSDWEGADMSYNGCLRGCNDSEDIGMEMOCMDTRINWAELRCENWNAYCSCOHYSHEDYHLMDMRNADNIAMELDOLMYDLGYLYNLYIHRMLCTINDARGTOAAYIGTEYWTDMSSO";
 
	    /*
	    The code above was generated as follows:

	    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=plug_board, reflector=reflector)
	    enigma.set_key([?,?,?])
	    enigma.set_plug([?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?])

	    cipher2 = enigma.encrypt('ALRIGHTTHISMESSAGEWAS?????????????????????????????????????????????????????????????????????????????????????????????????????????????????')

	    print cipher2
	    */

    
	    EnigmaEmulator machine = new EnigmaEmulator(alphabet, perm, plug_board, reflector);

	    int nplugs = 2;  
	    militaryEnigmaBreak(machine, nplugs, cipher2, "ALRIGHTTHISMESSAGEWAS");
    
    }
}
