#! /usr/bin/env python

import sys
import argparse
import time
import random


from EnigmaEmulator import *
from Numberjack import *

ITER = 0


def enumerate_reversed(L):
    for index in reversed(xrange(len(L))):
          yield index, L[index]


def BombeEnigmaBreak(ciphertext, enigma, crib, key, implied=False, bothways=False, subexpr=False): # data: a commercial enigma machine, a plaintext. Emulate the encoding
    N = enigma.alphabet.size()
    digit = enigma.alphabet.digit

    ForwardRotors = [VarArray([Variable(r,r,'fr_%i'%i) for i,r in enumerate(R.forward)]) for R in enigma.rotors]
    BackwardRotors = [VarArray([Variable(r,r,'br_%i'%i) for i,r in enumerate(R.backward)]) for R in enigma.rotors]
    
    PlugBoard = VarArray([Variable(0, N-1,'p_%i'%i) for i in range(N)])
    # PlugBoard = VarArray([Variable(i, i,'p_%i'%i) for i in range(N)])
    
    Reflector = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(enigma.reflector)])

    text = VarArray(len(crib), 0, N-1, 'text')
    plug_board_in = VarArray(len(crib), 0, N-1, 'pbin')
    plug_board_out = VarArray(len(crib), 0, N-1, 'pbout')
    reflected = VarArray(len(crib), 0, N-1, 'reflected')
    forward_cipher = [VarArray(len(crib), 0, N-1, 'forward_%i'%i) for i in range(len(enigma.rotors))]
    backward_cipher = [VarArray(len(crib), 0, N-1, 'backward_%i'%i) for i in range(len(enigma.rotors))]
    # key_var = VarArray(len(enigma.rotors), 0, N-1, 'key')

    model = Model(
        AllDiff(PlugBoard),
        Sum([PlugBoard[i]==i for i in range(N)]) == 6,
        [xt == digit[t] for xt,t in zip(plug_board_out,ciphertext)],
        [xt == digit[t] for xt,t in zip(text,crib)]
    )
    
    # for i in range(len(crib)):
    #     model.add( plug_board_in[i] != backward_cipher[0][i] )
    #     model.add( forward_cipher[0][i] != backward_cipher[1][i] )
    #     model.add( forward_cipher[1][i] != backward_cipher[2][i] )
    #     model.add( forward_cipher[2][i] != reflected[i] )

    if implied:
        for i in range(len(text)):
            for j in range(len(text)):
                if i<j:
                    if crib[i] == crib[j]:
                        model.add( plug_board_in[i] == plug_board_in[j] )
                    else:
                        model.add( plug_board_in[i] != plug_board_in[j] )
                    
                    if ciphertext[i] == ciphertext[j]:
                        model.add( backward_cipher[0][j] == backward_cipher[0][i] )
                    else:
                        model.add( backward_cipher[0][j] != backward_cipher[0][i] )
                    
                    if crib[i] == ciphertext[j]:
                        model.add( plug_board_in[i] == backward_cipher[0][j] )
                    else:
                        model.add( plug_board_in[i] != backward_cipher[0][j] )
                    
                    if ciphertext[i] == crib[j]:
                        model.add( backward_cipher[0][i] == plug_board_in[j] )
                    else:
                        model.add( backward_cipher[0][i] != plug_board_in[j] )
                 
    
    model.add( [x_next == PlugBoard[x_prev] for i,(x_prev,x_next) in enumerate(zip(text,plug_board_in))] )
    
    if bothways:
        model.add( [x_prev == PlugBoard[x_next] for i,(x_prev,x_next) in enumerate(zip(text,plug_board_in))] )
    
                
    previous = plug_board_in
    for j,(k,F,B,cipher) in enumerate(zip(key,ForwardRotors,BackwardRotors,forward_cipher)):
        model.add( [x_next == F[(x_prev + k + ((i//(N**j)) % N)) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        if bothways:
            model.add( [x_prev == (((B[x_next] + 2*N - ((i//(N**j)) % N)) - k) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
   
    model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

    previous = reflected
    for j,(k,F,B,cipher) in enumerate_reversed(zip(key,ForwardRotors,BackwardRotors,backward_cipher)):
        model.add( [x_prev == F[(k + x_next + ((i//(N**j)) % N)) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        if bothways:
            model.add( [x_next == (((B[x_prev] + 2*N - ((i//(N**j)) % N)) - k) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
        previous = cipher
    
    model.add( [x_next == PlugBoard[x_prev] for i,(x_prev,x_next) in enumerate(zip(previous,plug_board_out))] )
    
    if bothways:
        model.add( [x_prev == PlugBoard[x_next] for i,(x_prev,x_next) in enumerate(zip(previous,plug_board_out))] )


    search_vars = [p for p in PlugBoard]
    solver = model.load("Mistral2", search_vars) 
    solver.setVerbosity(0)
    
    ok = solver.solve()
    while ok:
        enigma.set_key(key)
        enigma.set_plugs([p.get_value() for p in PlugBoard])
        print enigma.encrypt(ciphertext)

        ok = solver.getNextSolution()



    
    
      
     
def bombe(enigma, ciphertext, crib):
    m = enigma
    k = len(m.rotors)
    N = m.alphabet.size()
    digit = m.alphabet.digit
    
    crib_t = [(i,t,c) for i, (t,c) in enumerate(zip(crib,ciphertext))]
    
    crib_dict = [[] for i in range(N)]
    crib_set = set([])
    for i,t,c in crib_t:
        crib_dict[digit[t]].append((i,digit[c]))
        crib_dict[digit[c]].append((i,digit[t]))
        crib_set.add(digit[c])
        crib_set.add(digit[t])

    m.set_key([0 for r in range(k)])
    m.plug_board = range(N)
    
    nkeys = 0
    niteration = N**k
    for iteration in xrange(niteration):
        if iteration % 100 == 0 :
            sys.stdout.write('\b\b\b\b\b\b\b\b\b%-5i %2i%%'%(nkeys, (iteration*100)//niteration))
            sys.stdout.flush()
        
        ikey = [(iteration//(N**i)) % N for i in range(len(m.rotors))]
        m.set_key(ikey)
        m.plug_board = range(N)

        forbidden = [[False]*N for i in range(N)]
        possible = [N]*N
        diagonal = N
        support = True  
    
        for t in crib_set:
            for s in m.alphabet.char:
                
                s = digit[s] 
                stack = [(t,s)]
                if t != s:
                    stack.append((s,t))
                spos = 0
                stecker = [None]*N
                while len(stack) > spos and not forbidden[t][s]:
                    x,y = stack[spos]
                    spos += 1
                    if not forbidden[x][y]:
                        stecker[x] = y
                        stecker[y] = x                
                        for p,cx in crib_dict[x]:

                            m.set_key([(k + p//(N**i)) % N for i,k in enumerate(ikey)])
                            ey = m.encryption(y)

                            if stecker[ey] is None and stecker[cx] is None:
                                stecker[ey] = cx
                                stecker[cx] = ey
                                stack.append((cx,ey))
                                if ey != cx :
                                    stack.append((ey,cx))
                            elif stecker[cx] != ey:
                                stack.append((cx,ey))
                                if ey != cx :
                                    stack.append((ey,cx))
                                for c1, c2 in stack:
                                    if not forbidden[c1][c2]:
                                        forbidden[c1][c2] = True
                                        possible[c1] -= 1
                                        if possible[c1] == 0:
                                            support = False
                                            
                                        if c1 == c2:
                                            diagonal -= 1
                                            if diagonal < 6:
                                                support = False
                                break
                if not support:
                    break
            if not support:
                break     
        if support:
            nkeys += 1            
            print '\b\b\b\b\b\b\b\b\btry key @%i'%iteration, ikey
            BombeEnigmaBreak(ciphertext, enigma, crib, ikey, implied=True, bothways=True)
            
    print 



  
        
if __name__ == '__main__':
    
    
    alphabet = Alphabet()
    
    cipher = 'CAUBOJJOUTCTVHXQFJZUBZQCNVIGFBUTJXSSZJRIWTPEKDJPZCVNEPQMCABQVZOYJRJJACYVJJGWFZNZTFLCEIGGHCNBXYGMUQHUYXVHTDDQLVCOKTHXCZSVSPGWTGAAGQVWKWOD'
    
    
    perm = [[0, 9, 20, 19, 7, 6, 4, 11, 3, 14, 22, 8, 17, 15, 2, 16, 23, 18, 10, 1, 12, 21, 24, 13, 5, 25], \
            [20, 15, 5, 12, 25, 14, 16, 2, 22, 21, 7, 9, 1, 24, 4, 19, 3, 11, 0, 18, 13, 10, 8, 17, 23, 6], \
            [13, 0, 2, 4, 14, 19, 16, 22, 25, 1, 7, 9, 17, 10, 3, 21, 15, 5, 24, 11, 12, 8, 20, 23, 18, 6]]
    reflector = [21, 13, 12, 14, 5, 4, 18, 25, 22, 23, 17, 24, 2, 1, 3, 16, 15, 10, 6, 20, 19, 0, 8, 9, 11, 7]
    

    
    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], reflector=reflector, plug_board=range(alphabet.size()))
    bombe(enigma, cipher, 'THEFIRSTONEWHOCAN')
  


    
