import random
import time


from Numberjack import *
from EnigmaEmulator import *
from Alphabet import *





def enumerate_reversed(L):
    for index in reversed(xrange(len(L))):
          yield index, L[index]
        
        
def militaryEnigmaBreak(alphabet, ciphertext, enigma, crib, nplugs, implied=False, bothways=False, subexpr=False): # data: a commercial enigma machine, a plaintext. Emulate the encoding
    N = alphabet.size()

    ForwardRotors = [VarArray([Variable(r,r,'fr_%i'%i) for i,r in enumerate(R.forward)]) for R in enigma.rotors]
    BackwardRotors = [VarArray([Variable(r,r,'br_%i'%i) for i,r in enumerate(R.backward)]) for R in enigma.rotors]
    
    PlugBoard = VarArray([Variable(0, N-1,'p_%i'%i) for i in range(N)])
    # PlugBoard = VarArray([Variable(i, i,'p_%i'%i) for i in range(N)])
    
    Reflector = VarArray([Variable(r,r,'r_%i'%i) for i,r in enumerate(enigma.reflector)])

    text = VarArray(len(crib), 0, N-1, 'text')
    plug_board_in = VarArray(len(crib), 0, N-1, 'pbin')
    plug_board_out = VarArray(len(crib), 0, N-1, 'pbout')
    reflected = VarArray(len(crib), 0, N-1, 'reflected')
    forward_cipher = [VarArray(len(crib), 0, N-1, 'forward_%i'%i) for i in range(len(enigma.rotors))]
    backward_cipher = [VarArray(len(crib), 0, N-1, 'backward_%i'%i) for i in range(len(enigma.rotors))]
    key_var = VarArray(len(enigma.rotors), 0, N-1, 'key')

    model = Model(
        AllDiff(PlugBoard),
        Sum([PlugBoard[i]==i for i in range(N)]) == N-2*nplugs,
        [xt == alphabet.digit[t] for xt,t in zip(plug_board_out,ciphertext)],
        [xt == alphabet.digit[t] for xt,t in zip(text,crib)]
    )
    
    # for i in range(len(crib)):
    #     model.add( plug_board_in[i] != backward_cipher[0][i] )
    #     model.add( forward_cipher[0][i] != backward_cipher[1][i] )
    #     model.add( forward_cipher[1][i] != backward_cipher[2][i] )
    #     model.add( forward_cipher[2][i] != reflected[i] )

    if implied:
        for i in range(len(text)):
            for j in range(len(text)):
                if i<j:
                    if crib[i] == crib[j]:
                        model.add( plug_board_in[i] == plug_board_in[j] )
                    else:
                        model.add( plug_board_in[i] != plug_board_in[j] )
                    
                    if ciphertext[i] == ciphertext[j]:
                        model.add( backward_cipher[0][j] == backward_cipher[0][i] )
                    else:
                        model.add( backward_cipher[0][j] != backward_cipher[0][i] )
                    
                    if crib[i] == ciphertext[j]:
                        model.add( plug_board_in[i] == backward_cipher[0][j] )
                    else:
                        model.add( plug_board_in[i] != backward_cipher[0][j] )
                    
                    if ciphertext[i] == crib[j]:
                        model.add( backward_cipher[0][i] == plug_board_in[j] )
                    else:
                        model.add( backward_cipher[0][i] != plug_board_in[j] )
                
                
    
    
    model.add( [x_next == PlugBoard[x_prev] for i,(x_prev,x_next) in enumerate(zip(text,plug_board_in))] )
    
    if bothways:
        model.add( [x_prev == PlugBoard[x_next] for i,(x_prev,x_next) in enumerate(zip(text,plug_board_in))] )
    
    
    if subexpr:
        
        f = []
        for i in range(len(crib)):
            f.append([(k + ((i//(N**j)) % N)) for j,k in enumerate(key_var)])
        # f = Matrix(len(crib), len(key_var), 0, N-1)
        # print f
        # for i in range(len(crib)):
        #     for j,k in enumerate(key_var):
        #         model.add(f[i][j] == (k + ((i//(N**j)) % N)))
        b = []
        for i in range(len(crib)):
            b.append([((2*N - ((i//(N**j)) % N)) - k) for j,k in enumerate(key_var)])
            
    
        previous = plug_board_in
        for j,(k,F,B,cipher) in enumerate(zip(key_var,ForwardRotors,BackwardRotors,forward_cipher)):
            model.add( [x_next == F[(x_prev + f[i][j]) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            if bothways:
                model.add( [x_prev == ((B[x_next] + b[i][j]) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            previous = cipher
       
        model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

        previous = reflected
        for j,(k,F,B,cipher) in enumerate_reversed(zip(key_var,ForwardRotors,BackwardRotors,backward_cipher)):
            model.add( [x_prev == F[(x_next + f[i][j]) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            if bothways:
                model.add( [x_next == ((B[x_prev] + b[i][j]) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            previous = cipher
        
    else:
                
        previous = plug_board_in
        for j,(k,F,B,cipher) in enumerate(zip(key_var,ForwardRotors,BackwardRotors,forward_cipher)):
            model.add( [x_next == F[(x_prev + k + ((i//(N**j)) % N)) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            if bothways:
                model.add( [x_prev == (((B[x_next] + 2*N - ((i//(N**j)) % N)) - k) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            previous = cipher
       
        model.add( [x_next == Reflector[x_prev] for i,(x_prev,x_next) in enumerate(zip(forward_cipher[-1],reflected))] )

        previous = reflected
        for j,(k,F,B,cipher) in enumerate_reversed(zip(key_var,ForwardRotors,BackwardRotors,backward_cipher)):
            model.add( [x_prev == F[(k + x_next + ((i//(N**j)) % N)) % N] for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            if bothways:
                model.add( [x_next == (((B[x_prev] + 2*N - ((i//(N**j)) % N)) - k) % N) for i,(x_prev,x_next) in enumerate(zip(previous,cipher))] )
            previous = cipher
    
    model.add( [x_next == PlugBoard[x_prev] for i,(x_prev,x_next) in enumerate(zip(previous,plug_board_out))] )
    
    if bothways:
        model.add( [x_prev == PlugBoard[x_next] for i,(x_prev,x_next) in enumerate(zip(previous,plug_board_out))] )


    # search_vars = [p for p in PlugBoard]
    # search_vars.extend([k for k in key_var])

    solver = model.load("Mistral2") #, search_vars)
    solver.setVerbosity(2)
    
    possible_keys = []
    start = time.clock()
    ok = solver.solve()
    while ok:
        solution = [k.get_value() for k in key_var], [p.get_value() for p in PlugBoard]
        possible_keys.append(solution)
        print 'key = %8s, %8inds, %.*fs'%('/'.join([str(k) for k in solution]),solver.getNodes(),2,(time.clock()-start))
        ok = solver.getNextSolution()
    print 'total = %8inds, %.*fs'%(solver.getNodes(),2,(time.clock()-start))
    return possible_keys

           
            
if __name__ == '__main__':
    
    random.seed(12345)
    
    alphabet = Alphabet('ACDEGHILMNORSTWY')
    
    
    perm = ([13, 9, 8, 10, 7, 15, 12, 14, 1, 5, 2, 4, 3, 11, 0, 6], \
            [1, 5, 2, 3, 7, 11, 8, 13, 10, 4, 14, 6, 12, 15, 0, 9], \
            [9, 13, 1, 10, 14, 0, 3, 12, 8, 11, 4, 7, 5, 2, 6, 15])
    reflector = [4, 13, 5, 7, 0, 2, 15, 3, 9, 8, 12, 14, 10, 1, 11, 6]
    
    

    print '\n\nEXERCICE 1\n'
    cipher2 = 'RCSDRLWISTHIYMRLTSDWEGADMSYNGCLRGCNDSEDIGMEMOCMDTRINWAELRCENWNAYCSCOHYSHEDYHLMDMRNADNIAMELDOLMYDLGYLYNLYIHRMLCTINDARGTOAAYIGTEYWTDMSSO'
 
    """
    The code above was generated as follows:

    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=plug_board, reflector=reflector)
    enigma.set_key([?,?,?])
    enigma.set_plug([?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?])

    cipher2 = enigma.encrypt('ALRIGHTTHISMESSAGEWAS?????????????????????????????????????????????????????????????????????????????????????????????????????????????????')

    print cipher2
    
    """

    
    enigma = Machine(alphabet, rotors=[Rotor(p) for p in perm], plug_board=range(alphabet.size()), reflector=reflector)

    nplugs = 2    
    possible_keys = militaryEnigmaBreak(alphabet, cipher2, enigma, 'ALRIGHTTHISMESSAGEWAS', nplugs, implied=True, subexpr=False, bothways=False)
    
    for keys,plugs in possible_keys:
        enigma.set_key(keys)
        enigma.set_plugs(plugs)
        print 'try key %8s: %s'%('/'.join([str(k) for k in keys]), enigma.encrypt(cipher2))

    
    
    
    