import random

from Numberjack import *
from Alphabet import *


def transposeEncodeChar(alphabet,permutation):
    
    N = alphabet.size()
    P = VarArray([Variable(p,p,'p_%i'%i) for i,p in enumerate(permutation)])
    
    text = Variable(0,N-1,'text') 
    cipher = Variable(0,N-1,'cipher')
    
    model = Model(
        cipher == P[text]
    ) 
    
    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    ok = solver.solve()
    while ok:
        print alphabet.char[text.get_value()], '->', alphabet.char[cipher.get_value()]
        ok = solver.getNextSolution()
            
            
def transposeEncodeText(alphabet,plaintext,permutation):
    
    N = alphabet.size()
    P = VarArray([Variable(p,p,'p_%i'%i) for i,p in enumerate(permutation)])
    K = len(plaintext)
    
    ciphertext = VarArray(K,0,N-1,'text') 
    
    model = Model(
        [c == P[alphabet.digit[t]] for t,c in zip(plaintext,ciphertext)]
    ) 
    
    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    solver.solve()
    print plaintext, '->', ''.join([alphabet.char[c.get_value()] for c in ciphertext])

        
def transposeDecodeText(alphabet,ciphertext,permutation):
    
    N = alphabet.size()
    P = VarArray([Variable(p,p,'p_%i'%i) for i,p in enumerate(permutation)])
    K = len(ciphertext)
    
    plaintext = VarArray(K,0,N-1,'text') 
    
    model = Model(
        [Variable(alphabet.digit[c],alphabet.digit[c]) == P[t] for t,c in zip(plaintext,ciphertext)]
    ) 
    
    solver = model.load("Mistral2")
    solver.setVerbosity(0)
    solver.solve()
    print ''.join([alphabet.char[t.get_value()] for t in plaintext]), '->', ciphertext
            
            
if __name__ == '__main__':
    
    random.seed(12345)
    
    alphabet = Alphabet()
    
    permutation = [15, 14, 5, 22, 24, 13, 25, 12, 18, 9, 1, 23, 20, 16, 17, 21, 7, 2, 3, 11, 4, 8, 6, 19, 0, 10]
    
    print '\nEXERCICE 1\n'
    
    transposeEncodeChar(alphabet, permutation)
    
    print '\n\nEXERCICE 2\n'
    
    transposeEncodeText(alphabet, "IAMAREALSPYNOW", permutation)

    print '\n\nEXERCICE 3\n'
    
    transposeDecodeText(alphabet, "MYXXRPQWGYXFRUYLRLMYOECYPE", permutation)
    
    
    
    
    